#brailleDisplayDrivers/alvaSatellite.py
#A part of NonVisual Desktop Access (NVDA)
#Copyright (C) 2006-2009 NVDA Contributors <http://www.nvda-project.org/>
#This file is covered by the GNU General Public License.
#See the file COPYING for more details.

import os
import braille
import queueHandler
from logHandler import log
from ctypes import *
from ctypes.wintypes import *
import time
import wx
import config
import inputCore

ALVA_RELEASE_MASK = 0x80
ALVA_KEY_CHECK_INTERVAL = 50

# Key groups
ALVA_NO_KEY = 0x70 # No key available
ALVA_FRNT_GRP = 0x71 # front keys
ALVA_CR1_GRP = 0x72 # Lower row of routing corsor keys
ALVA_CR2_GRP = 0x75 # Upper row of routing cursor keys
ALVA_SAT1_GRP = 0x77 # Left satellite keypad
KEY_ERROR = 0xFF

# Logical Key Groups
ALVA_STA1_GRP = 0x81 # Lower row of status routing cursor
ALVA_STA2_GRP = 0x82 # Upper row of status routing cursor
ALVA_SAT2_GRP = 0x87 # Right satellite keypad.

ALVA_KEYS = {
	# FRONT GROUP
	ALVA_FRNT_GRP: ("prog", "home", "cursor", "up", "left",
                        "right", "down", "cursor2", "home2", "prog2"),
	# Left Satellite pad
	ALVA_SAT1_GRP: ("lf1", "lup", "lleft", "ldown", "lright", "lf2"),
	# Right Satellite pad
	ALVA_SAT2_GRP: ("rf1", "rup", "rleft", "rdown", "rright", "rf2"),
	# Lower status routing cursor
	ALVA_STA1_GRP: ("st1", "st2", "st3"),
	# Upper status routing cursor
	ALVA_STA2_GRP: ("st4", "st5", "st6"),
}

# List of supported Alva Display Terminals
ALVA_TYPE = [ "Unsupported", "Unsupported", "Unsupported", "ABT", "Delphi", "Satellite", "Unsupported", "Unsupported" ]

# Try to load ABT.dll
try:
	#AlvaLib = windll[os.path.abspath(os.path.join(os.path.dirname(__file__), r"ABT.dll"))]
	AlvaLib = windll[os.getcwdu() + r"\\brailleDisplayDrivers\\" + r"ABT.dll"]
	AlvaLib.ABT_GetABTType.restype = c_byte*4
	AlvaLib.ABT_GetKeyStatus.restype = c_byte*4
except:
	AlvaLib=None

class BrailleDisplayDriver(braille.BrailleDisplayDriver):
	name = "alvaABT"
	# Translators: The name of a braille display.
	description = _("ALVA ABT/Delphi/Satellite series")

	@classmethod
	def check(cls):
		if AlvaLib:
			log.info("Alva library was loaded")
		else:
			log.info("Alva library was not loaded")
		return bool(AlvaLib)

	def __init__(self):
		super(BrailleDisplayDriver,self).__init__()
		log.debug("ALVA ABT/Delphi/Satellite Braille init")
		if not AlvaLib.ABT_InstallABTport(""):
			raise RuntimeError("No ALVA display found")
		AlvaLib.ABT_Clear()
		self._alva_NumCells = 0
		self._keysDown = set()
		self._ignoreKeyReleases = False
		self._alva_KeyCheckTimer = wx.PyTimer(self._alva_CheckKeyPresses)
		self._alva_KeyCheckTimer.Start(ALVA_KEY_CHECK_INTERVAL)

	def terminate(self):
		super(BrailleDisplayDriver, self).terminate()
		AlvaLib.ABT_Clear()
		try:
			self._alva_KeyCheckTimer.Stop()
			self._alva_KeyCheckTimer = None
		except:
			pass

	def _get_numCells(self):
		if self._alva_NumCells==0:
                        type = AlvaLib.ABT_GetABTType()
                        log.info("ALVA %s %d%d device found, %d cells, %d status cells" %
                                 (ALVA_TYPE[type[1]], type[1], type[0], type[2], type[3]))
                        self._alva_NumCells = type[2]
		return self._alva_NumCells

	def display(self, cells):
		log.debug("Display %d cells" %len(cells))
		cells="".join([chr(x) for x in cells])
		AlvaLib.ABT_BrDisplay(cells, 0, len(cells))

	def _alva_CheckKeyPresses(self):
		while True:
			try:
				key_status = AlvaLib.ABT_GetKeyStatus()
				group = key_status[1]
				key_code = key_status[2]
				if group == ALVA_NO_KEY:
					break
                                if key_code & ALVA_RELEASE_MASK:
                                        is_key_release = True
                                        key_code = key_code & 0x7f
                                else:
                                        is_key_release = False
                                if (group == ALVA_FRNT_GRP) and (key_code >= 0x30):
                                        key_code = key_code - 0x30
                                        group = ALVA_STA2_GRP
                                if (group == ALVA_FRNT_GRP) and (key_code >= 0x20):
                                        key_code = key_code - 0x20
                                        group = ALVA_STA1_GRP
                                if (group == ALVA_SAT1_GRP) and (key_code >= 0x20):
                                        key_code = key_code - 0x20
                                        group = ALVA_SAT2_GRP
                                log.debug("Key: Released=%d, group=%d, code=%d" % (is_key_release, group, key_code))
			except:
				log.error("Error reading keypress from ABT.DLL", exc_info=True)
				return
			if is_key_release:
                                # Release.
                                if not self._ignoreKeyReleases and self._keysDown:
                                        try:
                                                inputCore.manager.executeGesture(InputGesture(self._keysDown))
                                        except inputCore.NoInputGestureAction:
                                                pass
                                        # Any further releases are just the rest of the keys in the combination being released,
                                        # so they should be ignored.
                                        self._ignoreKeyReleases = True
                                self._keysDown.discard((group, key_code))
			else:
                                # Press.
                                if group == ALVA_CR1_GRP:
                                        # Execute routing keys when pressed instead of released.
                                        try:
                                                inputCore.manager.executeGesture(InputGesture(((group, key_code),)))
                                        except inputCore.NoInputGestureAction:
                                                pass
                                else:
                                        self._keysDown.add((group, key_code))
                                        # This begins a new key combination.
                                        self._ignoreKeyReleases = False

	gestureMap = inputCore.GlobalGestureMap({
		"globalCommands.GlobalCommands": {
			"braille_scrollBack": ("br(alvaABT):left",),
			"braille_previousLine": ("br(alvaABT):up",),
			"braille_nextLine": ("br(alvaABT):down",),
			"braille_scrollForward": ("br(alvaABT):right",),
			"braille_routeTo": ("br(alvaABT):routing",),
			"braille_toggleTether": ("br(alvaABT):cursor+home",),
			"navigatorObject_toFocus": ("br(alvaABT):cursor",),
			"review_top": ("br(alvaABT):home", "br(alvaABT):up+left",),
			"review_bottom": ("br(alvaABT):down+right",),
			"review_startOfLine": ("br(alvaABT):home+left",),
			"review_endOfLine": ("br(alvaABT):home+right",),
			"navigatorObject_moveToFlatReviewAtObjectPosition": ("br(alvaABT):home+up",),
			"navigatorObject_moveToObjectAtFlatReviewPosition": ("br(alvaABT):home+down",),
			"kb:shift+tab": ("br(alvaABT):rf1",),
			"kb:alt": ("br(alvaABT):lf1",),
			"kb:escape": ("br(alvaABT):lf2",),
			"kb:tab": ("br(alvaABT):rf2",),
			"kb:upArrow": ("br(alvaABT):rup",),
			"kb:downArrow": ("br(alvaABT):rdown",),
			"kb:leftArrow": ("br(alvaABT):rleft",),
			"kb:rightArrow": ("br(alvaABT):rright",),
			"kb:enter": ("br(alvaABT):rf1+rf2",),
			"kb:windows+m": ("br(alvaABT):ldown",),
			"showGui": ("br(alvaABT):lleft",),
			"kb:windows+d": ("br(alvaABT):lright",),
			"kb:windows": ("br(alvaABT):lup",),
			"kb:alt+tab": ("br(alvaABT):lf1+lf2",),
		}
	})

class InputGesture(braille.BrailleDisplayGesture):
	source = BrailleDisplayDriver.name
	def __init__(self, keys):
		super(InputGesture, self).__init__()
		self.keyCodes = set(keys)
		self.keyNames = names = set()
		for group, number in self.keyCodes:
			if (group == ALVA_CR1_GRP) or (group == ALVA_CR2_GRP):
				names.add("routing")
				self.routingIndex = number
			else:
				try:
					names.add(ALVA_KEYS[group][number])
				except (KeyError, IndexError):
					log.debugWarning("Unknown key with group %d and number %d" % (group, number))
		self.id = "+".join(names)
